import scipy as sp
import numpy as np
import math
from sklearn.neighbors import kneighbors_graph
import igraph as ig
import umap

from .Spectral import spectral
from .Clustering import leiden

def reduceDimension(args):
    spectral(args)

"""
args.input: a list of files.
args.k: number of neighbours.
args.output: output file.
args.embed: embeding file.
"""
def mkKNNGraph(args):
    fls = args.input.split(',')
    adj = None

    #adj = kneighbors_graph(readCoordinates(fls[0]), args.k, mode='distance', n_jobs=12)
    for fl in fls:
        mat = readCoordinates(fl)
        if (adj == None):
            adj = kneighbors_graph(mat, args.k, mode='distance', n_jobs=args.thread)
        else:
            adj += kneighbors_graph(mat, args.k, mode='distance', n_jobs=args.thread)
    adj = adj / len(fls)

    np.reciprocal(adj.data, out=adj.data)
    sp.sparse.save_npz(args.output, adj)

    if(args.embed):
        print("Create Embedding:")
        data = readCoordinates(fls[0])
        getEmbedding(data, args.embed)

def clustering(args):
    adj = sp.sparse.load_npz(args.input)
    vcount = max(adj.shape)
    sources, targets = adj.nonzero()
    if (args.perturb):
        toDelete = set()
        for i in range(vcount):
            _, col = adj.getrow(i).nonzero()
            idx = np.random.choice(col)
            toDelete.add((i, idx))
            toDelete.add((idx, i))
        print("Perturbing ")
        edges = filter(lambda x: x not in toDelete, zip(sources.tolist(), targets.tolist()))
        sources, targets = zip(*edges)
    edgelist = list(zip(list(sources), list(targets)))
    weights = np.ravel(adj[(sources, targets)])
    gr = ig.Graph(n=vcount, edges=edgelist, edge_attrs={"weight": weights})

    print("Start Clustering...")
    partition = leiden(gr, resolution=args.res, optimizer=args.optimizer, seed=args.seed)

    n = 0
    with open(args.output, 'w') as f:
        cutoff = max(1, min(50, 0.001*vcount))
        for c in partition:
            if len(c) > cutoff:
                n = n + 1
                print(','.join(map(str, c)), file=f) 
    print(n)

def readCoordinates(fl, n_dim=None, discard=None, scale=None):
    def scaling(xs):
        s = 0
        for x in xs:
            s = s + x * x
        s = math.sqrt(s)
        return np.array([x / s for x in xs])
    data = np.loadtxt(fl)
    if (n_dim):
        data = data[..., :n_dim]
    if (discard):
        data = data[..., 1:]
    if (scale):
        data = np.apply_along_axis(scaling, 1, data)
    return data

def getEmbedding(mat, output):
    embedding = umap.UMAP(random_state=42, n_components=2).fit_transform(mat)
    np.savetxt(output, embedding, delimiter='\t')




import scrublet as scr
import scipy as sp
import numpy as np
from sklearn.mixture import BayesianGaussianMixture

from .Utils import readMatrix

def detectDoublet(args):
    counts_matrix = readMatrix(args.input, binary=False)
    scrub = scr.Scrublet(counts_matrix,
        expected_doublet_rate = 0.06, sim_doublet_ratio=3, n_neighbors=25)
    doublet_scores, _ = scrub.scrub_doublets(
        min_counts=1, 
        min_cells=3, 
        min_gene_variability_pctl=85,
        mean_center=True,                                                                                                                                                                                                                                                                
        normalize_variance=True,
        n_prin_comps=30
        )

    # Fit a Gaussian mixture model
    X = scrub.doublet_scores_sim_
    X = np.array([X]).T
    gmm = BayesianGaussianMixture(n_components=2, max_iter=1000,
        random_state=2394).fit(X)
    i = np.argmax(gmm.means_)

    probs_sim = gmm.predict_proba(X)[:,i]
    vals = X[np.argwhere(probs_sim>0.5)].flatten()
    threshold = min(vals)

    X = np.array([doublet_scores]).T
    probs = gmm.predict_proba(X)[:,i].tolist()

    with open(args.output, 'w') as fl:
        fl.write('\t'.join(map(str, probs)))
        fl.write("\n")

        fl.write(str(threshold))
        fl.write("\n")
        fl.write('\t'.join(map(str, (doublet_scores.tolist()))))
        fl.write("\n")
        fl.write('\t'.join(map(str, scrub.doublet_scores_sim_)))


import scipy as sp
import numpy as np
import math
from sklearn.linear_model import LinearRegression
from sklearn.metrics.pairwise import cosine_similarity

from .Utils import readMatrix, regress

def spectral(args):
    nSample = max(1000, args.sample_size)
    nChunk = nSample

    print("Read Data")
    if (args.distance == "jaccard"):
        mat = readMatrix(args.input, binary=True)
    else:
        mat = readMatrix(args.input, binary=False)

    n, _ = mat.get_shape()
    if nSample < n:
        np.random.seed(args.seed)
        idx = np.arange(n)
        np.random.shuffle(idx)
        sample = mat[idx[:nSample], :]
        dm = Spectral(sample, n_dim=args.dim, distance=args.distance, sampling_rate=nSample/n)
        i = nSample
        res = [dm.coordinates]
        while i < n:
            data = mat[idx[i:i+nChunk], :]
            res.append(dm.fit(data))
            i = i + nChunk
        res = np.concatenate(res, axis=0)[:, 1:]
        res = res[np.argsort(idx), :]
    else:
        res = Spectral(mat, n_dim=args.dim, distance=args.distance).coordinates[:, 1:]

    np.savetxt(args.output, res, delimiter='\t')
    
class Spectral:
    def __init__(self, mat, n_dim=30, sampling_rate=1, distance="jaccard"):
        self.sample = mat
        self.sampling_rate = sampling_rate
        self.dim = mat.get_shape()[1]
        self.coverage = mat.sum(axis=1) / self.dim
        self.distance = distance
        if (self.distance == "jaccard"):
            print("Use jaccard distance")
            self.compute_similarity = jaccard_similarity
        else:
            self.compute_similarity = cosine_similarity

        print("Compute similarity matrix")
        jm = self.compute_similarity(mat)

        if (self.distance == "jaccard"):
            self.normalizer = Normalizer(jm, self.coverage)
            S = self.normalizer.fit(jm, self.coverage, self.coverage)
        else:
            S = jm

        np.fill_diagonal(S, 0)
        print("Normalization")
        self.D = np.diag(1/(self.sampling_rate * S.sum(axis=1)))
        L = np.matmul(self.D, S)

        print("Reduction")
        evals, evecs = sp.sparse.linalg.eigs(L, n_dim+1, which='LR')
        ix = evals.argsort()[::-1]
        self.evals = np.real(evals[ix])
        self.evecs = np.real(evecs[:, ix])
        self.coordinates = self.evecs

    def fit(self, data):
        jm = self.compute_similarity(self.sample, data)

        if (self.distance == "jaccard"):
            S_ = self.normalizer.fit(jm, self.coverage, data.sum(axis=1) / self.dim).T
        else:
            S_ = jm.T

        D_ = np.diag(1/(self.sampling_rate * S_.sum(axis=1)))
        L_ = np.matmul(D_, S_)
        evecs = (L_.dot(self.evecs)).dot(np.diag(1/self.evals))
        return evecs

class Normalizer:
    def __init__(self, jm, c):
        n, _ = jm.shape

        X = 1 / c.dot(np.ones((1,n)))
        X = 1 / (X + X.T - 1)
        X = X[np.triu_indices(n, k = 1)].T
        y = jm[np.triu_indices(n, k = 1)].T

        self.model = LinearRegression().fit(X, y)

    def fit(self, jm, c1, c2):
        X1 = 1 / c1.dot(np.ones((1, c2.shape[1]))) 
        X2 = 1 / c2.dot(np.ones((1, c1.shape[1])))
        X = 1 / (X1 + X2.T - 1)

        y = self.model.predict(X.flatten().T).reshape(jm.shape)
        return np.array(jm / y)

# Similarity metric
""" Compute pair-wise jaccard index
Input:
    mat1: n1 x m
    mat2: n2 x m
Output:
    jm: n1 x n2
"""
def jaccard_similarity(mat1, mat2=None):
    coverage1 = mat1.sum(axis=1)
    if(mat2 != None):
        coverage2 = mat2.sum(axis=1)
        jm = mat1.dot(mat2.T).todense()
        n1, n2 = jm.shape
        c1 = coverage1.dot(np.ones((1,n2)))
        c2 = coverage2.dot(np.ones((1,n1)))
        jm = jm / (c1 + c2.T - jm)
    else:
        n, _ = mat1.get_shape()
        jm = mat1.dot(mat1.T).todense()
        c = coverage1.dot(np.ones((1,n)))
        jm = jm / (c + c.T - jm)
    return jm


import scipy as sp
import numpy as np
import math
import leidenalg as la

def leiden(gr, resolution=1, optimizer="RB", seed=12343):
    weights = gr.es["weight"]
    if (optimizer == "RB"):
        algo = la.RBConfigurationVertexPartition
    elif (optimizer == "CPM"):
        algo = la.CPMVertexPartition
    partition = la.find_partition(gr, algo,
        n_iterations=10, seed=seed, resolution_parameter=resolution,
        weights=weights)
    return partition


import numpy as np
from sklearn.neighbors import KDTree
import itertools
from sklearn.cluster import KMeans

class Projector(object):
    def __init__(self, X, Y):
        self.reference = X
        self.vector = Y - X

    def project(self, X, weight=0.5):
        def project(x):
            P = self.reference
            U = self.vector
            d = np.sqrt(np.sum((P - x)**2, axis=1))
            w = np.exp(-(d/0.005))
            #w = 1/d
            return (x + weight * np.average(U, axis=0, weights=w))
        return np.apply_along_axis(project, 1, X)

def findMNC(X, Y, k, n):
    cX = KMeans(n_clusters=k, random_state=0).fit(X).cluster_centers_
    cY = KMeans(n_clusters=k, random_state=0).fit(Y).cluster_centers_

    treeX = KDTree(cX)
    treeY = KDTree(cY)

    # X by Y matrix
    mX = treeY.query(cX, k=n, return_distance=False)

    # Y by X matrix
    mY_ = treeX.query(cY, k=n, return_distance=False)
    mY = []
    for i in range(mY_.shape[0]):
        mY.append(set(mY_[i,:]))

    iX = []
    iY = []
    for i in range(mX.shape[0]):
        for j in mX[i]:
            if i in mY[j]:
                iX.append(i)
                iY.append(j)
    a = cX[iX,:]
    b = cY[iY,:]
    return (Projector(a, b), Projector(b, a))

def MNCCorrect(dat, k, n):
    X = dat[0]
    for i in range(1, len(dat)):
        Y = dat[i]
        pX, pY = findMNC(X, Y, k, n)
        ratio = X.shape[0] / (X.shape[0] + Y.shape[0])
        X_ = pX.project(X, weight = 1 - ratio)
        Y_ = pY.project(Y, weight = ratio)
        X = np.concatenate((X_, Y_), axis=0)
    return X

def MNCCorrectMain(args):
    mat = np.loadtxt(args.input)
    with open(args.label, 'r') as fl:
        labels = [l.strip() for l in fl]
    
    for _ in range(args.iter):
        batchLabels = list(set(labels))
        batchIndex = []
        batch = []
        for label in batchLabels:
            idx = [i for i, x in enumerate(labels) if x == label]
            batchIndex.append(idx)
            batch.append(mat[idx,:])

        mat_ = MNCCorrect(batch, args.k, args.n)
        new = np.array(mat, copy=True)
        idx = list(itertools.chain.from_iterable(batchIndex))
        for i in range(len(labels)):
            new[idx[i]] = mat_[i,:]
        mat = new
    np.savetxt(args.output, mat, delimiter='\t')


import argparse
import taiji_utils as sc

from .Doublet import detectDoublet
from .BatchCorrect import MNCCorrectMain

################################################################################
## ARGUMENT PARSER
################################################################################

parser = argparse.ArgumentParser(description="Fast online algorithms")
subparsers = parser.add_subparsers(title="sub commands")

# create the parser for the "reduce" command
parser_reduce = subparsers.add_parser('reduce', help='dimension reduction')
parser_reduce.add_argument('input', type=str, help='gzipped input file')
parser_reduce.add_argument('output', type=str, help='output matrix in .npy format')
parser_reduce.add_argument('--sample-size', default=35000, type=int, help='sampling size')
parser_reduce.add_argument('--dim', default=30, type=int, help='number of dimension')
parser_reduce.add_argument('--seed', default=3484, type=int, help='random seed')
parser_reduce.add_argument('--distance', default="jaccard", type=str, help='distance: jaccard or cosine')
parser_reduce.set_defaults(func=sc.reduceDimension)

# create the parser for the "knn" command
parser_knn = subparsers.add_parser('knn', help='Make KNN graph')
parser_knn.add_argument('input', type=str, help='input matrix in .npy format')
parser_knn.add_argument('output', type=str, help='adjacency matrix')
parser_knn.add_argument('-k', default=50, type=int, help='neighbors')
parser_knn.add_argument('--embed', help='embedding file')
parser_knn.add_argument('--thread', default=1, type=int, help='number of jobs')
parser_knn.set_defaults(func=sc.mkKNNGraph)

# create the parser for the "clust" command
parser_clust = subparsers.add_parser('clust', help='perform clustering')
parser_clust.add_argument('input', type=str, help='adjacency matrix')
parser_clust.add_argument('output', type=str, help='output file')
parser_clust.add_argument('--res', type=float, default=1, help='resolution')
parser_clust.add_argument('--perturb', type=float, help='perturb')
parser_clust.add_argument('--seed', type=int, default=12343, help='seed')
parser_clust.add_argument('--optimizer', type=str, default="RB", help='algorithm: RB, CPM')
parser_clust.set_defaults(func=sc.clustering)

# create the parser for the "doublet" command
parser_doublet = subparsers.add_parser('doublet', help='doublet detection')
parser_doublet.add_argument('input', type=str, help='input matrix')
parser_doublet.add_argument('output', type=str, help='output')
parser_doublet.set_defaults(func=detectDoublet)

# create the parser for the "correct" command
parser_correct = subparsers.add_parser('correct', help='batch correction')
parser_correct.add_argument('input', type=str, help='input matrix')
parser_correct.add_argument('output', type=str, help='output')
parser_correct.add_argument('--label', type=str, help='labels')
parser_correct.add_argument('-k', type=int, default=20, help='number of centroids')
parser_correct.add_argument('-n', type=int, default=2, help='number of nearest neighbors')
parser_correct.add_argument('--iter', type=int, default=1, help='number of iterations')
parser_correct.set_defaults(func=MNCCorrectMain)

def main():
    args = parser.parse_args()
    args.func(args)


import gzip
import scipy as sp
import numpy as np
from sklearn.linear_model import LinearRegression

class InputData:
    def __init__(self, filename):
        self.filename = filename
        with gzip.open(self.filename, mode='rt') as f:
            header = f.readline().strip()
            (m, n) = header.split(": ")[1].split(" x ")
            self.num_doc = int(m)
            self.num_terms = int(n)

    def __iter__(self):
        def convert(x):
            return (int(x[0]), float(x[1]))
        with gzip.open(self.filename, mode='rt') as f:
            f.readline()
            for line in f:
                yield [convert(item.split(",")) for item in line.strip().split("\t")[1:]]

# regress out a variable
def regress(X, y):
    model = LinearRegression().fit(X, y)
    return model.predict(X)

def readMatrix(fl, binary=False):
    data = InputData(fl)
    indptr = [0]
    indices = []
    mat = []
    for row in iter(data):
        for (i,x) in row:
            indices.append(i)
            if (binary):
                mat.append(1)
            else:
                mat.append(x)
        indptr.append(len(indices))
    mat = sp.sparse.csr_matrix((mat, indices, indptr), shape=(data.num_doc, data.num_terms), dtype=int)
    return mat


from setuptools import setup

setup(name='taiji-utils',
      version='0.1.2',
      description='Taiji pipeline',
      url='https://taiji-pipeline.github.io/', 
      author='Kai Zhang',
      author_email='kai@kzhang.org',
      license='MIT',
      packages=['taiji_utils'],
      entry_points = {
        'console_scripts': ['taiji-utils=taiji_utils.command_line:main'],
      },
      install_requires=[
          'scipy',
          'numpy',
          'scikit-learn',
          'python-igraph',
          'umap-learn',
          'leidenalg',
          'scrublet',
      ],
      zip_safe=False)